%
%   =====================================================================
%   Copyright (C) 2019  
%   Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   Stefan Schubert, stefan.schubert@etit.tu-chemnitz.de
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% Neubert, P., Schubert, S. & Protzel, P. (2019) A neurologically inspired 
% sequence processing model for mobile robot place recognition. In IEEE 
% Robotics and Automation Letters (RA-L) and presentation at Intl. Conf. 
% on Intelligent Robots and Systems (IROS). DOI: 10.1109/LRA.2019.2927096 
%
%
% 
% This class provides the main MCN functionality. Starting with an empty 
% set of minicolumns, each time a feedforward pattern is seen that is not
% similar to any existing minicolumn, create a new one witch responds to
% inputs similar to this pattern. 
% 
% For details, please refer to the paper. To see how MCN is used to create a
% similarity matrix from two descriptor sequences, please see run_MCN.m
% For a full demonstration see demo_MCN.m
% 
classdef MCN < handle

properties
    % parameter-struct for MCN. Must contain:
    %
    % params.probAdditionalCon ... probability for random connection
    % params.nCellPerCol ... number of cells per minicolumn
    % params.nInConPerCol ... number of connections per minicolumn
    % params.minColumnActivity ... minicolumn activation threshold θ
    % params.nColsPerPattern ... minimum number of active minicolumns k_min
    % params.kActiveColumn ... maximum number of active minicolumns k_max
    params
    
    % feedforward connections: FF(:,j) is the set of indices into the input
    % SDR of the j-th minicolumn. So the activation of all minicolumns can
    % be computed by sum(SDR(FF))
    FF
    P % P(i,j) is the predicted-flag for the i-th cell in the j-th minicolumn
    prevP
        
    nCols % current number of minicolumns in the network
    
    burstedCol  % flag if this minicolumn bursts
    prevBurstedCol
    
    winnerCells % idx of all winning cells (in matrix of same size as P)
    activeCells % idx of all active cells (in matrix of same size as P)
    
    prevWinnerCells
    prevActiveCells
    
    % datastructure that holds the connections for predictions
    predictionConnections 
    predictionConnections_backward
    
    % flag to supress learning and switch some behaviours (e.g., how the
    % active columns are generated)
    supressLearningFlag
end
   
methods
    function obj = MCN(params)
        obj.params = params;
        
        obj.P = [];
        obj.prevP = [];
        
        obj.burstedCol = [];
        obj.prevBurstedCol = [];
        
        obj.params.nReserveCols = 5000;
        obj.predictionConnections = {};
        obj.predictionConnections_backward = {};
        obj.reserveStorage();
        
        obj.nCols = 0;                
    end
    
    % reset some internal variables
    function prepareNewIteration(obj)
        obj.prevWinnerCells = obj.winnerCells;
        obj.winnerCells = [];
        obj.prevActiveCells = obj.activeCells;
        obj.activeCells = [];
        
        obj.prevP = obj.P;
        obj.prevBurstedCol = obj.burstedCol;        
        if obj.nCols>0
            obj.P(:) = 0;
            obj.burstedCol(:) = 0;
        end
    end
    
    % inputConnections ... idx in a potential input SDR ,stored in obj.FF
    % return index of the ne column
    function newColIdx = createNewColumn(obj, inputConnections, nNewCols)
        for i=1:nNewCols
            obj.nCols = obj.nCols+1; 
            if obj.nCols > size(obj.FF,2)
                reserveStorage(obj);
            end
            
            obj.FF(:,obj.nCols) = obj.datasample(inputConnections, obj.params.nInConPerCol)';        
            obj.P(:,obj.nCols) = zeros(obj.params.nCellPerCol, 1);     
            obj.prevP(:,obj.nCols) = zeros(obj.params.nCellPerCol, 1);
            obj.predictionConnections(:, obj.nCols) = cell(obj.params.nCellPerCol, 1);
            obj.predictionConnections_backward(:, obj.nCols) = cell(obj.params.nCellPerCol, 1);
            obj.burstedCol(:,obj.nCols) = 0;
            obj.prevBurstedCol(:,obj.nCols) = 0;
        end      
        newColIdx = obj.nCols-nNewCols+1 : obj.nCols;
    end
    
    function reserveStorage(obj)
        k = obj.params.nReserveCols;
        obj.FF(:,end+1:end+k) = zeros(obj.params.nInConPerCol, k);        
        obj.P(:,end+1:end+k) = zeros(obj.params.nCellPerCol, k);     
        obj.prevP(:,end+1:end+k) = zeros(obj.params.nCellPerCol, k);
        obj.predictionConnections(:, end+1:end+k) = cell(obj.params.nCellPerCol, k);
        obj.predictionConnections_backward(:, end+1:end+k) = cell(obj.params.nCellPerCol, k);
        obj.burstedCol(:,end+1:end+k) = 0;
        obj.prevBurstedCol(:,end+1:end+k) = 0;
    end
    
    % - Compare inputSDR to all minicolumns to find active minicolumns
    % - Search for predicted cells in active minicolumns and activate their
    %   predictions
    function compute(obj, inputSDR, supressLearningFlag)
        if ~exist('supressLearningFlag', 'var')
            supressLearningFlag = 0;
            obj.supressLearningFlag = 0;
        else
            obj.supressLearningFlag = supressLearningFlag;
        end
        
        obj.prepareNewIteration();
        
        % How similar is the input SDR to the pattern of the minicolumns?        
        columnActivity = obj.computeColumnActivations(inputSDR);
        [columnActivity_sorted, idx_sorted] = sort(columnActivity, 'descend');
        
        if ~supressLearningFlag
            % Are there activities above threshold? If yes, activate the k most
            % active columns, otherwise create new ones and make these the active ones.
            if ~isempty(columnActivity)
              activeCols = idx_sorted(columnActivity_sorted>obj.params.minColumnActivity ...
                                      & columnActivity_sorted>=columnActivity_sorted(min(obj.params.kActiveColumn, length(columnActivity_sorted))));
            else
              activeCols = [];
            end
            
            sdrNonZeroIdx = find(inputSDR);
            activeCols = [activeCols, obj.createNewColumn(sdrNonZeroIdx, max(0, obj.params.nColsPerPattern - length(activeCols)))];   
        else
            % In non-learning mode, take the k most active columns
            % plus columns with same activity like kth-best column           
            activeCols = idx_sorted(columnActivity_sorted>=columnActivity_sorted(min(obj.params.kActiveColumn, length(columnActivity_sorted))));
        end       
        
        % for each active column:
        % - mark all predicted cells as winnerCells
        % - if there was no predicted cell, chose one and activate all
        % predictions
        % - activate predictions of winnerCells
        %
        
        obj.activeCells = [];
        obj.winnerCells = [];

        for i=activeCols
            predictedIdx = find(obj.prevP(:,i)>0); 
            
            if isempty(predictedIdx)
                % if there are no predicted: burst (predict from all cells
                % and choose one winner cell)                
                winnerCell = obj.burst(i);
                obj.winnerCells = [obj.winnerCells, ...
                    sub2ind(size(obj.P), winnerCell, repmat(i, 1, length(winnerCell)))];
                obj.activeCells = [obj.activeCells, ...
                    sub2ind(size(obj.P), 1:size(obj.P,1), repmat(i,1,size(obj.P,1)))];
            elseif length(predictedIdx)==1
                % if there is only one predicted cell, make this the
                % winner cell
                winnerCell = predictedIdx;
                obj.activatePredictions(i, winnerCell);                
                obj.winnerCells = [obj.winnerCells, ...
                    sub2ind(size(obj.P), winnerCell, i)];              
                obj.activeCells = [obj.activeCells, ...
                    sub2ind(size(obj.P), winnerCell, i)];
            else
                % if there are multiple predicted cells, make all winner
                % cells (according to Marcus)   
                for j=predictedIdx'                    
                    obj.activatePredictions(i,j);
                    obj.winnerCells = [obj.winnerCells, ...
                    sub2ind(size(obj.P), j, i)];  
                    obj.activeCells = [obj.activeCells, ...
                        sub2ind(size(obj.P), j, i)];
                end
            end              
        end
        
        % learn predictions
        if ~supressLearningFlag
            obj.learnPredictions();
        
            % also predict newly learned predictions
            for columnIdx=1:obj.nCols
                if obj.burstedCol(columnIdx) == 1        
                    for i=1:size(obj.P,1)
                        obj.activatePredictions(columnIdx, i);
                    end
                end
            end
        end        
    end

    % given the set of currently winning cells and the set of previously
    % winning cells, create prediction connection 
    function learnPredictions(obj)
        for curIdx = obj.winnerCells
          
            for prevIdx = obj.prevWinnerCells
                 
                [~, prevColIdx] = ind2sub(size(obj.P), prevIdx);

                % check whether the previous column is already connected to
                % the cell
                existingPredConFlag = obj.checkExistingPredCon(prevColIdx, curIdx);
                if ~existingPredConFlag || rand(1)<=obj.params.probAdditionalCon
                    obj.predictionConnections_backward{curIdx}(end+1) = prevIdx;                       
                    obj.predictionConnections{prevIdx}(:, end+1) = curIdx;
                end
            end  
            
            % check for identical connections if random connections are
            % enabled
            if obj.params.probAdditionalCon ~= 0
              if ~isempty(obj.predictionConnections_backward{curIdx}) % otherwise a 0x1 matrix is created      
                  obj.predictionConnections_backward{curIdx} = unique(obj.predictionConnections_backward{curIdx});
              end
            end
        end    
    end
    
    % Check if there already is an predicting connection from the previous
    % column to this active cell. This is used during bursting to prevent
    % learning multiple connections from one column to a single cell. In
    % this case, the new connection should go to a new cell of the current
    % column, to indicate the different context.
    function existingPredConFlag = checkExistingPredCon(obj, prevColIdx, curCellIdx)
      [~, prevColIdxs] = ind2sub(size(obj.P), obj.predictionConnections_backward{curCellIdx});      
      existingPredConFlag = any(prevColIdxs==prevColIdx);      
    end
    
    % Activate the predictions of all cells and identify the cell with the
    % fewest forward predictions to be the winning cell. winnerCellIdx is the 
    % index of this cell in the minicolumn
    function winnerCellIdx = burst(obj, columnIdx)
        obj.burstedCol(columnIdx) = 1;
        
        for i=1:size(obj.P,1)
            obj.activatePredictions(columnIdx, i);
        end
        
        % winnerCell: one of the cells with fewest existing forward predictions   
        nForwardPredictionsPerCell = cellfun(@length, obj.predictionConnections(:,columnIdx));
        
        if ~obj.supressLearningFlag
            % (slightly) inhibit winning cells from the last iteration a little bit
            for i=1:length(obj.prevWinnerCells)
                [cellIdx, colIdx] = ind2sub(size(obj.P), obj.prevWinnerCells(i));
                if colIdx==columnIdx
                    nForwardPredictionsPerCell(cellIdx) = nForwardPredictionsPerCell(cellIdx) + obj.params.nCellPerCol;
                end
            end

            candidateIdx = find(nForwardPredictionsPerCell==min(nForwardPredictionsPerCell));
            winnerCellIdx = obj.resolveTie(candidateIdx);
        else
            % in case of inference, return all used winner cells
            winnerCellIdx = find(nForwardPredictionsPerCell>0)';
            if isempty(winnerCellIdx)
                winnerCellIdx = randi(obj.params.nCellPerCol);
            end
        end
    end
    
    % randomly select one element to break a tie
    function idx = resolveTie(obj, x)
        idx = obj.datasample(x,1);
    end
    
    function y = datasample(obj, x, n)
        ii = randi(numel(x),[1, n]);
        y = x(ii);   
    end
    
    % Increase number of predictions for all cells that are predicted from this
    % cell
    function activatePredictions(obj, colIdx, cellIdx)
        predIdx = obj.predictionConnections{cellIdx, colIdx};
        obj.P(predIdx) = obj.P(predIdx) + 1;      
    end        
        
    % evaluate all connections between the input space and columns
    function columnActivity = computeColumnActivations(obj, inputSDR)
        columnActivity = sum(inputSDR(obj.FF(:, 1:obj.nCols)));
    end
  
end

end % class
