%
%   =====================================================================
%   Copyright (C) 2019  
%   Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   Stefan Schubert, stefan.schubert@etit.tu-chemnitz.de
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% Neubert, P., Schubert, S. & Protzel, P. (2019) A neurologically inspired 
% sequence processing model for mobile robot place recognition. In IEEE 
% Robotics and Automation Letters (RA-L) and presentation at Intl. Conf. 
% on Intelligent Robots and Systems (IROS). DOI: 10.1109/LRA.2019.2927096 
%
%
% Demonstrates MCN and pairwise matching on a small example dataset.
%
function demo_MCN()

  %% set random seed for repeatability
  rng(0);

  %% load descriptors
  D1 = load('data/summer_alexnet_lsh.mat');
  D1 = normr(D1.D);
  D2 = load('data/winter_alexnet_lsh.mat');
  D2 = normr(D2.D);

  %% sLSBH (sparse Locality Sensitive Binary Hashing) - binarize descriptors
  P = randn(8192, 16384, 'single');
  P = normc(P); 
  
  D1_slsbh = getLSBH(D1, P, 0.25);
  D2_slsbh = getLSBH(D2, P, 0.25);
  
  clear('P');

  %% experiments
  % pairwise (raw descriptors)
  S_pairwise = 1 - pdist2(D1, D2, 'cosine');
  
  % sLSBH (binarized descriptors)
  S_sLSBH = evaluateSimilarity(D1_slsbh, D2_slsbh);
  
  % MCN (binary descriptors with sequence information)
  S_MCN = run_MCN(D1_slsbh, D2_slsbh);

  %% start evaluation
  figure; hold all;
  
  % ground truth
  GT = eye(288);
  
  % pairwise (raw)
  [P, R] = createPR(S_pairwise, GT);
  plot(R, P);
  legends{1} = ['pairwise / raw (avgP=', num2str(trapz(R, P)), ')'];
  
  % sLSBH
  [P, R] = createPR(S_sLSBH, GT);
  plot(R, P);
  legends{2} = ['sLSBH / binarized (avgP=', num2str(trapz(R, P)), ')'];
  
  % MCN
  [P, R] = createPR(S_MCN, GT);
  plot(R, P);
  legends{3} = ['MCN (avgP=', num2str(trapz(R, P)), ')'];

  % visualization
  legend(legends, 'location', 'southwest');
  xlabel('Recall');
  ylabel('Precision');
  grid('on');
  title('Place recognition performance comparison');
end
