%
%   =====================================================================
%   Copyright (C) 2019  
%   Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   Stefan Schubert, stefan.schubert@etit.tu-chemnitz.de
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% Neubert, P., Schubert, S. & Protzel, P. (2019) A neurologically inspired 
% sequence processing model for mobile robot place recognition. In IEEE 
% Robotics and Automation Letters (RA-L) and presentation at Intl. Conf. 
% on Intelligent Robots and Systems (IROS). DOI: 10.1109/LRA.2019.2927096 
%
%
% Evaluate pairwise similarity of SDRs in D1 and D2 (this is used similar
% to 1-pdist2(D1, D2) but for sparse vectors).
%
% D1,D2 ... either a measurement/encoding data matrix with each row being a
% a sample (e.g. a descriptor) OR a cell array with each entry beeing the
% vector or indices of ones in a row
%
function S = evaluateSimilarity(D1, D2)
    
    if nargin<2
        D2 = D1;
        symetricFlag = true;
    else
        symetricFlag = false;
    end

    if iscell(D1)        
        % convert to sparse
        max_cell_id_D1 = max(cellfun(@max, D1));
        if ~symetricFlag
          max_cell_id_D2 = max(cellfun(@max, D2));
          max_cell_id = max(max_cell_id_D1, max_cell_id_D2);
          D2_sparse = sparse(length(D2), max_cell_id);
          
          for i = 1:length(D2)
              D2_sparse(i, D2{i}) = 1;
          end
          D2 = D2_sparse;
          
        else
          max_cell_id = max_cell_id_D1;
        end

        D1_sparse = sparse(length(D1), max_cell_id);
        for i = 1:length(D1)
            D1_sparse(i, D1{i}) = 1;
        end
        D1 = D1_sparse;
    end
    
    % copy D1 if symmetric
    if symetricFlag
      D2 = D1;
    end
    
    % pairwise comparison   
    S = D1 * D2';
    nOnes_D1 = sum(D1, 2);
    nOnes_D2 = sum(D2, 2);
    mean_nOnes = [ones(length(nOnes_D1), 1), nOnes_D1] * [nOnes_D2'; ones(1, length(nOnes_D2))] / 2;
    
    S = S ./ mean_nOnes;
    
    if issparse(S)
      S = full(S);
    end
    
end
