%
%   =====================================================================
%   Copyright (C) 2019  
%   Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   Stefan Schubert, stefan.schubert@etit.tu-chemnitz.de
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% Neubert, P., Schubert, S. & Protzel, P. (2019) A neurologically inspired 
% sequence processing model for mobile robot place recognition. In IEEE 
% Robotics and Automation Letters (RA-L) and presentation at Intl. Conf. 
% on Intelligent Robots and Systems (IROS). DOI: 10.1109/LRA.2019.2927096 
%
%
% Apply MCN to create a similarity matrix from two descriptor sequences.
%
% D1/D2 ... descriptor matrices, each row is a sparse binary descriptor (e.g. 
% an sLSBH encoding)
% params_in ... MCN parameter struct, optional (default parameters can be
% found below)
%
function S = run_MCN(D1, D2, params_in)
   
  %% parameters
    
  % default values
  params.probAdditionalCon = 0.05; % probability for random connection
  params.nCellPerCol = 32; % number of cells per minicolumn
  params.nInConPerCol = 200; % number of connections per minicolumn
  params.minColumnActivity = 0.75; % minicolumn activation threshold θ
  params.nColsPerPattern = 50; % minimum number of active minicolumns k_min
  params.kActiveColumn = 100; % maximum number of active minicolumns k_max

  % use params from input
  if nargin > 3
    for field = fieldnames(params_in)'
      params.(field{1}) = params_in.(field{1});
    end
  end

  % conversion of the parameter to a natural number that contains the
  % required number of 1s for activation
  params.minColumnActivity = round(params.minColumnActivity*params.nInConPerCol);
    
  %% initialize MCN
  mcn = MCN(params);
  
  %% run MCN on D1
  outputSDR1 = cell(size(D1,1), 1);

  fprintf('-------- D1 ITERATION ---------\n');
  for i=1:size(D1,1)
    if mod(i,100)==0
      fprintf('-------- D1 ITERATION %d/%d (nCols=%d) ---------\n', i, size(D1,1), mcn.nCols);
    end
    mcn.compute(D1(i,:));
    outputSDR1{i} = mcn.winnerCells;   
  end

  %% run MCN on D2
  outputSDR2 = cell(size(D2,1), 1);

  fprintf('-------- D2 ITERATION ---------\n');
  for i=1:size(D2,1)
    if mod(i,100)==0
      fprintf('-------- D2 ITERATION %d/%d (nCols=%d) ---------\n', i, size(D2,1), mcn.nCols);
    end
    mcn.compute(D2(i,:), 1); % suppress learning
    outputSDR2{i} = mcn.winnerCells;    
  end
    
  %% compute similarity matrix
  S = evaluateSimilarity(outputSDR1, outputSDR2);
end


